//DevMpf.h

/*************************************************************************
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory, and the Regents of the University of California, as
* Operator of Los Alamos National Laboratory. MPF is distributed
* subject to a Software License Agreement found in file LICENSE that
* is included with this distribution.
*************************************************************************/
/*
	Current Author: Marty Kraimer
	Original Author: Jim Kowalkowski (for Hideos)
	Date: 01JUL98

	This is the main base class for any EPICS MPF device support.  All
	EPICS device support that communicates with MPF tasks should be
	derived from this class.  This class hides most of the details of EPICS
	device support and the MPF (Message Passing Facility).

	Each instance of a device is bound to one MPF server, the name of the
	MPF task it is bound to is the first string in parm field of the 
	record's link.
*/

 
#ifndef DevMpfH
#define DevMpfH

class Message;
class MessageClient;

#include <epicsEvent.h>
#include <epicsMutex.h>
#include <callback.h>
#include <dbCommon.h>
#include <link.h>
#include <dbScan.h>
#include <devSup.h>
#include <epicsExport.h>

#include "Message.h"
#include "WatchDog.h"
#include "ConnectMessage.h"
#include "OutOfBandMessage.h"
/*------------------------------------------------------------------------*/
/* generic structure for device support */
extern "C" {
typedef struct {
	long        number;
	DEVSUPFUN   report;
	DEVSUPFUN   init;
	DEVSUPFUN   init_record;
	long (*get_ioint_info)(int,dbCommon*,IOSCANPVT*);
	DEVSUPFUN   read_write;
	long (*conv)(void*,int);
} MPF_DSET;
}

/* Macros that generate DSETs */

#define MAKE_DSET(name,init_func) \
extern "C"\
{\
static long name##myinit_func (void *pvt) \
    {return(init_func(pvt));} \
static long name##ioint (int pass,dbCommon *pdb, IOSCANPVT *iosp) \
    {return(DevMpf::ioint(pass,pdb,iosp));} \
static long name##read_write (void *pvt) \
    {return(DevMpf::read_write(pvt));} \
static MPF_DSET name = \
  { 5,NULL,NULL,name##myinit_func,name##ioint,name##read_write,NULL };\
epicsExportAddress(dset,name); \
}

#define MAKE_LINCONV_DSET(name,init_func) \
extern "C" \
{\
static long name##myinit_func (void *pvt) \
    {return(init_func(pvt));} \
static long name##ioint (int pass,dbCommon *pdb, IOSCANPVT *iosp) \
    {return(DevMpf::ioint(pass,pdb,iosp));} \
static long name##read_write (void *pvt) \
    {return(DevMpf::read_write(pvt));} \
static long name##linconv (void *pvt,int pass) \
    {return(DevMpf::linconv(pvt,pass));} \
static MPF_DSET name = \
 { 6,NULL,NULL,name##myinit_func,name##ioint,name##read_write, name##linconv};\
epicsExportAddress(dset,name); \
}

/*------------------------------------------------------------------------*/
// return codes for startIO().

#define MPF_OK			0
#define MPF_NoConvert		2

enum replyType {replyTypeNone, replyTypeCompleteIO, replyTypeReceiveReply};
enum waitState {
    waitStateNone,
    waitStateConnect,
    waitStateReply,
    waitStateTimeout
};

class DevMpf
{
public:
    DevMpf(dbCommon*,DBLINK*,bool iointValid);
    virtual ~DevMpf();
    // Following must be implemented by device support modules
    virtual long startIO(dbCommon*)=0;	// start async IO
    virtual long completeIO(dbCommon*,Message *)=0;	// end async IO
    // Following can be implemented by device support modules
    virtual void receiveReply(dbCommon*,Message *); //
    virtual void connectIO(dbCommon*,Message *);	// connection message
    virtual void outOfBandIO(dbCommon*,Message *);// outOfBand message
    virtual long convert(dbCommon*,int pass);	// do linear conversion
    // The following must be called after object is constructed
    void bind();
    // The following are usefull during record initilization
    bool connectWait() {return(connectWait(5.0));}
    bool connectWait(double timeout);
    int send(Message *pmessage) {return(send(pmessage,replyTypeNone));}
    int sendReply(Message *pmessage)
        {return(send(pmessage,replyTypeCompleteIO));}
    // send a message stating the type of reply
    int send(Message*,replyType); 
    int sendWait(Message *pmessage,replyType type)
        {return(sendWait(pmessage,type,5.0));}
    int sendWait(Message*,replyType,double timeout);
    // This routine gets a pointer to the user portion of the parm field
    const char* getUserParm() const { return((const char*)userParm); }
    long getStatus() const { return(status);}
    // Following are DSET routines
    static long read_write(void*);	// generic DSET read/write routine
    static long ioint(int cmd,dbCommon*,IOSCANPVT* iopvt); // DSET i/o intr
    static long linconv(void*,int); // calls convert
    IOSCANPVT ioscanpvt;
    bool iointValid;
    bool isConnected() { return((connectState==connectYes) ? true : false);}
    //Following not normally needed by user code
    void   lock() {epicsMutexMustLock(lockId);}
    void   unlock() {epicsMutexUnlock(lockId);}
private:
    static void messageClientCallback(Message *message,void *clientPvt);
    static void sendWaitCallback(void *pvt);
    long status;
    dbCommon *precord;
    DBLINK  *plink;
    char* userParm;
    char* serverName;
    int serverLocation;
    CALLBACK cb;
    MessageClient *pMessageClient;
    connectStatus connectState;
    Message *replyMessage;
    waitState state;
    WatchDog sendWatchdog;
    epicsMutexId lockId;
    epicsEventId sendWaitSem;
    // disallow copy constructor, assignment operator
    DevMpf(const DevMpf &);
    DevMpf& operator=(const DevMpf&);
};

#endif // DevMpfH
